from minet.crawl.types import CrawlJob
from minet.crawl.url_cache import AtomicSet, SQLiteStringSet, URLCache


class TestUrlCache:
    def test_atomic_set(self):
        s = AtomicSet()

        assert len(s) == 0

        assert s.add("one")
        assert not s.add("one")

        assert len(s) == 1
        assert "one" in s
        assert "two" not in s

        assert s.add_many(["one", "two", "three"]) == 2

        assert len(s) == 3
        assert "two" in s

        assert set(s) == {"one", "two", "three"}

        assert s.add_many_and_keep_new(["one", "three", "four"]) == ["four"]

        assert len(s) == 4

        assert s.add_many_and_keep_new(
            [(0, "one"), (2, "five")], key=lambda t: t[1]
        ) == [(2, "five")]

    def test_sqlite_string_set(self, tmp_path):
        p = tmp_path

        s = SQLiteStringSet(p)

        assert len(s) == 0

        assert s.add("one")
        assert not s.add("one")

        assert len(s) == 1

        assert "one" in s
        assert "two" not in s

        # Testing persistence
        del s

        s = SQLiteStringSet(p)

        assert len(s) == 1
        assert "one" in s
        assert "two" not in s

        assert s.add_many(["one", "two", "three"]) == 2
        assert s.add_many(["one", "two", "three"]) == 0
        assert s.add_many(["four", "five"]) == 2

        assert len(s) == 5

        s.vacuum()

        assert len(s) == 5

        assert set(s) == {"one", "two", "three", "four", "five"}

        del s

        s = SQLiteStringSet(p)

        assert len(s) == 5

        assert s.add_many_and_keep_new(["one", "three", "six", "seven"]) == [
            "six",
            "seven",
        ]

        assert len(s) == 7

        assert s.add_many_and_keep_new(
            [(0, "one"), (2, "eight")], key=lambda t: t[1]
        ) == [(2, "eight")]

    def test_url_cache(self):
        c = URLCache()

        def jobs_as_urls(jobs):
            return [job.url for job in jobs]

        assert not c.persistent

        assert jobs_as_urls(c.register([CrawlJob(url="one"), CrawlJob(url="two")])) == [
            "one",
            "two",
        ]

        assert c.register([CrawlJob(url="one"), CrawlJob(url="two")]) == []

        assert len(c) == 2

        assert jobs_as_urls(
            c.register([CrawlJob(url="two"), CrawlJob(url="three")])
        ) == ["three"]
        assert len(c) == 3

        assert "two" in c
        assert "four" not in c

    def test_normalized_url_cache(self):
        c = URLCache(normalized=True)

        c.register([CrawlJob(url="lemonde.fr"), CrawlJob(url="www.lemonde.fr")])

        assert len(c) == 1

        assert "https://www.lemonde.fr/index.html" in c
        assert "https://www.lemonde.fr/articles/one.html" not in c



from typing import (
    TypeVar,
    Generic,
    Set,
    Iterable,
    Iterator,
    List,
    Optional,
    Callable,
    overload,
)

import sqlite3
from os import makedirs
from os.path import join
from threading import Lock
from contextlib import contextmanager
from ebbe import distinct
from ural import normalize_url

from minet.crawl.types import CrawlJob, CrawlJobDataType

T = TypeVar("T")
I = TypeVar("I")


class AtomicSet(Generic[T]):
    """
    A simple set wrapper ensuring all operations are
    atomic and thus threadsafe.
    """

    __items: Set[T]
    __lock: Lock

    def __init__(self):
        self.__items = set()
        self.__lock = Lock()

    def add(self, item: T) -> bool:
        with self.__lock:
            len_before = len(self.__items)

            self.__items.add(item)

            return len(self.__items) > len_before

    def add_many(self, items: Iterable[T]) -> int:
        with self.__lock:
            len_before = len(self.__items)

            for item in items:
                self.__items.add(item)

            return len(self.__items) - len_before

    @overload
    def add_many_and_keep_new(
        self, items: Iterable[I], key: Callable[[I], T] = ...
    ) -> List[I]:
        ...

    @overload
    def add_many_and_keep_new(self, items: Iterable[T], key: None = ...) -> List[T]:
        ...

    def add_many_and_keep_new(self, items, key=None):
        with self.__lock:
            new = []

            for item in items:
                k = item if key is None else key(item)

                len_before = len(self.__items)
                self.__items.add(k)

                if len_before < len(self.__items):
                    new.append(item)

            return new

    def __len__(self) -> int:
        with self.__lock:
            return len(self.__items)

    def __contains__(self, item: T) -> bool:
        with self.__lock:
            return item in self.__items

    def __iter__(self) -> Iterator[T]:
        with self.__lock:
            yield from self.__items


class SQLiteStringSet:
    def __init__(self, path: str, db_name: str = "set.db"):
        makedirs(path, exist_ok=True)

        self.path = path

        self.__connection = sqlite3.connect(
            join(self.path, db_name), check_same_thread=False
        )
        self.__lock = Lock()

        # Setup
        self.__connection.execute("PRAGMA journal_mode=wal;")
        self.__connection.execute(
            'CREATE TABLE IF NOT EXISTS "set" ("key" TEXT PRIMARY KEY);'
        )
        self.__connection.commit()

    @contextmanager
    def __transaction(self):
        try:
            with self.__lock, self.__connection:
                cursor = self.__connection.cursor()
                yield cursor
        finally:
            cursor.close()

    def vacuum(self) -> None:
        with self.__transaction() as cursor:
            cursor.execute("VACUUM;")

    def add(self, item: str) -> bool:
        with self.__transaction() as cursor:
            try:
                cursor.execute('INSERT INTO "set" ("key") VALUES (?);', (item,))
            except sqlite3.IntegrityError:
                return False

            return True

    def add_many(self, items: Iterable[str]) -> int:
        with self.__transaction() as cursor:
            rows = [(item,) for item in items]
            cursor.executemany('INSERT OR IGNORE INTO "set" ("key") VALUES (?);', rows)
            return cursor.rowcount

    @overload
    def add_many_and_keep_new(
        self, items: Iterable[I], key: Callable[[I], str] = ...
    ) -> List[I]:
        ...

    @overload
    def add_many_and_keep_new(self, items: Iterable[str], key: None = ...) -> List[str]:
        ...

    def add_many_and_keep_new(self, items, key=None):
        with self.__transaction() as cursor:
            new = []

            for item in items:
                try:
                    k = item if key is None else key(item)
                    cursor.execute('INSERT INTO "set" ("key") VALUES (?);', (k,))
                    new.append(item)
                except sqlite3.IntegrityError:
                    pass

            return new

    # NOTE: I keep this for the ugly trick.
    # def contains_many(self, items: Iterable[str]) -> List[str]:
    #     with self.__transaction() as cursor:
    #         cursor.execute(
    #             'SELECT "key" FROM "set" WHERE "key" in (SELECT "value" FROM JSON_EACH(?));',
    #             (json.dumps(items),),
    #         )

    def __contains__(self, item: str) -> bool:
        with self.__transaction() as cursor:
            cursor.execute('SELECT 1 FROM "set" WHERE "key" = ?;', (item,))
            return cursor.fetchone() is not None

    def __len__(self) -> int:
        with self.__transaction() as cursor:
            cursor.execute('SELECT COUNT(*) FROM "set";')
            return cursor.fetchone()[0]

    def __iter__(self) -> Iterator[str]:
        with self.__transaction() as cursor:
            cursor.execute('SELECT "key" FROM "set";')

            while True:
                rows = cursor.fetchmany(128)

                if not rows:
                    return

                for row in rows:
                    yield row[0]

    def __del__(self) -> None:
        self.__connection.close()


# TODO: should canonicalize by default
class URLCache:
    def __init__(self, path: Optional[str] = None, normalized: bool = False):
        self.path = path
        self.persistent = path is not None
        self.normalized = normalized

        self.__cache = (
            SQLiteStringSet(path, "urls.db") if path is not None else AtomicSet()
        )

    def register(
        self, jobs: Iterable[CrawlJob[CrawlJobDataType]]
    ) -> List[CrawlJob[CrawlJobDataType]]:
        def url_key(job: CrawlJob[CrawlJobDataType]):
            if self.normalized:
                return normalize_url(job.url)

            return job.url

        # We deduplicate beforehand
        jobs = distinct(jobs, key=url_key)
        new = self.__cache.add_many_and_keep_new(jobs, key=url_key)

        return new

    def __len__(self) -> int:
        return len(self.__cache)

    def __contains__(self, url: str) -> bool:
        if self.normalized:
            url = normalize_url(url)

        return url in self.__cache



# =============================================================================
# Minet CrowdTangle Links Summary
# =============================================================================
#
# Function related to link summary using CrowdTangle's APIs.
#
from ebbe import getpath
from urllib.parse import quote

from minet.crowdtangle.exceptions import CrowdTangleMissingTokenError
from minet.crowdtangle.constants import (
    CROWDTANGLE_SUMMARY_DEFAULT_SORT_TYPE,
    CROWDTANGLE_SUMMARY_SORT_TYPES,
)
from minet.crowdtangle.formatters import format_post, format_summary

URL_TEMPLATE = (
    "https://api.crowdtangle.com/links"
    "?token=%(token)s"
    "&count=%(count)s"
    "&startDate=%(start_date)s"
    "&includeSummary=true"
    "&link=%(link)s"
    "&sortBy=%(sort_by)s"
)


def url_forge(link, token, start_date, sort_by, platforms=None, include_posts=False):
    base_url = URL_TEMPLATE % {
        "token": token,
        "count": 1 if not include_posts else 1000,
        "start_date": start_date,
        "link": quote(link, safe=""),
        "sort_by": sort_by,
    }

    if platforms:
        base_url += "&platforms=%s" % ",".join(platforms)

    return base_url


def crowdtangle_summary(
    request,
    link,
    token=None,
    start_date=None,
    with_top_posts=False,
    sort_by=CROWDTANGLE_SUMMARY_DEFAULT_SORT_TYPE,
    raw=False,
    platforms=None,
):
    if token is None:
        raise CrowdTangleMissingTokenError

    if not isinstance(start_date, str):
        raise TypeError("minet.crowdtangle.summary: expecting a `start_date` kwarg.")

    if sort_by not in CROWDTANGLE_SUMMARY_SORT_TYPES:
        raise TypeError("minet.crowdtangle.summary: unknown `sort_by`.")

    # Fetching
    api_url = url_forge(link, token, start_date, sort_by, platforms, with_top_posts)

    data = request(api_url)

    stats = getpath(data, ["summary", "facebook"])
    posts = getpath(data, ["posts"]) if with_top_posts else None

    if stats is not None:
        if not raw:
            stats = format_summary(stats)

    if not with_top_posts:
        return stats

    else:
        if not raw:
            posts = [format_post(post, link=link) for post in posts]

        return stats, posts



# =============================================================================
# Minet BuzzSumo Exceptions
# =============================================================================
#
from minet.exceptions import MinetError


class BuzzSumoError(MinetError):
    pass


class BuzzSumoInvalidTokenError(BuzzSumoError):
    pass


class BuzzSumoInvalidQueryError(BuzzSumoError):
    def __init__(self, msg=None, url=None, data=None):
        super().__init__(msg)
        self.url = url
        self.data = data


class BuzzSumoBadRequestError(BuzzSumoError):
    pass


class BuzzSumoOutageError(BuzzSumoError):
    pass



from typing import TypeVar, Union, Generic, Optional

from nanoid import generate
from ebbe import format_repr
from functools import partial
from ural import get_domain_name

from minet.web import Response
from minet.serialization import serialize_error_as_slug

CrawlJobDataType = TypeVar("CrawlJobDataType")
CrawlResultDataType = TypeVar("CrawlResultDataType")


class CrawlTarget(Generic[CrawlJobDataType]):
    """
    This class represent a crawl target, i.e. at least a URL, and should be
    used by end-users when they need to specify what they want to crawl more
    precisely (i.e. if they need to switch spiders, have a custom depth or
    rely on some necessary data that will be passed along).
    """

    __slots__ = ("url", "depth", "spider", "data")

    url: str
    depth: Optional[int]
    spider: Optional[str]
    data: Optional[CrawlJobDataType]

    def __init__(
        self,
        url: str,
        depth: Optional[int] = None,
        spider: Optional[str] = None,
        data: Optional[CrawlJobDataType] = None,
    ) -> None:
        if not isinstance(url, str):
            raise TypeError("url should be a string")

        self.url = url

        if depth is not None and not isinstance(depth, int):
            raise TypeError("depth should be an int")

        self.depth = depth

        if spider is not None and not isinstance(spider, str):
            raise TypeError("spider should be a string")

        self.spider = spider
        self.data = data

    def __eq__(self, other):
        return (
            self.url == other.url
            and self.depth == other.depth
            and self.spider == other.spider
            and self.data == other.data
        )

    def __repr__(self):
        return format_repr(self, conditionals=("spider", "data", "depth"))


UrlOrCrawlTarget = Union[str, CrawlTarget[CrawlJobDataType]]


generate_crawl_job_id = partial(
    generate,
    alphabet="0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz",
    size=12,
)


class CrawlJob(Generic[CrawlJobDataType]):
    """
    This class represents an actual job that will be queued by the crawler. It
    should not be created by users themselves as they are inferred from
    a CrawlTarget instead.

    This class should be pickle-safe as it will be serialized by persistent
    queues.
    """

    __slots__ = (
        "id",
        "url",
        "depth",
        "spider",
        "data",
        "attempts",
        "parent",
        "__has_cached_domain",
        "__domain",
    )

    id: str
    url: str
    depth: int
    spider: Optional[str]
    data: Optional[CrawlJobDataType]
    attempts: int
    parent: Optional[str]

    # TODO: we should add headers, cookies and such here in the future

    __has_cached_domain: bool
    __domain: Optional[str]

    def __init__(
        self,
        url: str,
        depth: Optional[int] = None,
        spider: Optional[str] = None,
        data: Optional[CrawlJobDataType] = None,
        parent: Optional[str] = None,
    ):
        self.id = generate_crawl_job_id()
        self.url = url
        self.depth = depth if depth is not None else 0
        self.spider = spider
        self.data = data
        self.attempts = 0
        self.parent = parent

        self.__has_cached_domain = False
        self.__domain = None

    def __hash__(self) -> int:
        return hash(self.id)

    def __getstate__(self):
        return (
            self.id,
            self.url,
            self.depth,
            self.spider,
            self.data,
            self.attempts,
            self.parent,
        )

    def __setstate__(self, state):
        (
            self.id,
            self.url,
            self.depth,
            self.spider,
            self.data,
            self.attempts,
            self.parent,
        ) = state

        self.__has_cached_domain = False
        self.__domain = None

    @property
    def domain(self) -> Optional[str]:
        if self.__has_cached_domain:
            return self.__domain

        if self.url is not None:
            self.__domain = get_domain_name(self.url)

        self.__has_cached_domain = True

        return self.__domain

    def __repr__(self):
        return format_repr(self, conditionals=("data", "spider", "parent"))


class CrawlResult(Generic[CrawlJobDataType, CrawlResultDataType]):
    __slots__ = ("job", "data", "error", "response", "degree")

    FIELDNAMES = [
        "id",
        "parent",
        "spider",
        "depth",
        "url",
        "resolved_url",
        "error",
        "status",
        "degree",
        "body_size",
    ]

    job: CrawlJob[CrawlJobDataType]
    data: Optional[CrawlResultDataType]
    error: Optional[Exception]
    response: Optional[Response]
    degree: int

    def __init__(self, job: CrawlJob[CrawlJobDataType]):
        self.job = job
        self.data = None
        self.error = None
        self.response = None
        self.degree = 0

    @property
    def url(self) -> str:
        return self.job.url

    @property
    def depth(self) -> int:
        return self.job.depth

    @property
    def spider(self) -> Optional[str]:
        return self.job.spider

    @property
    def error_code(self) -> Optional[str]:
        return serialize_error_as_slug(self.error) if self.error else None

    def as_csv_row(self):
        job = self.job

        return [
            job.id,
            job.parent,
            job.spider,
            job.depth,
            job.url,
            self.response.end_url if self.response else None,
            self.error_code,
            self.response.status if self.response else None,
            self.degree,
            len(self.response) if self.response else None,
        ]

    def __repr__(self) -> str:
        return format_repr(
            self,
            attributes=(
                (
                    "url",
                    "depth",
                    "spider",
                    ("error", self.error_code),
                    ("status", self.response.status if self.response else None),
                    "degree",
                    (
                        "dtype",
                        type(self.data).__name__ if self.data is not None else None,
                    ),
                    ("size", self.response.human_size if self.response else None),
                )
            ),
            conditionals=("spider", "error", "status", "degree", "dtype"),
        )


class ErroredCrawlResult(CrawlResult[CrawlJobDataType, None]):
    job: CrawlJob[CrawlJobDataType]
    data: None
    error: Exception
    response: Optional[Response]
    degree: int

    def __init__(
        self,
        job: CrawlJob[CrawlJobDataType],
        error: Exception,
        response: Optional[Response] = None,
    ):
        self.job = job
        self.data = None
        self.error = error
        self.response = response
        self.degree = 0

    @property
    def error_code(self) -> str:
        return serialize_error_as_slug(self.error)


class SuccessfulCrawlResult(CrawlResult[CrawlJobDataType, CrawlResultDataType]):
    job: CrawlJob[CrawlJobDataType]
    data: CrawlResultDataType
    error: None
    response: Response
    degree: int

    def __init__(
        self,
        job: CrawlJob[CrawlJobDataType],
        response: Response,
        data: CrawlResultDataType,
        degree: int,
    ):
        self.job = job
        self.data = data
        self.error = None
        self.response = response
        self.degree = degree

    @property
    def error_code(self) -> None:
        return None


AnyCrawlResult = Union[
    ErroredCrawlResult[CrawlJobDataType],
    SuccessfulCrawlResult[CrawlJobDataType, CrawlResultDataType],
]



# =============================================================================
# Minet YouTube Formatters
# =============================================================================
#
# Various data formatters for YouTube data.
#
from casanova import namedrecord
from ebbe import getpath

from minet.youtube.constants import (
    YOUTUBE_VIDEO_CSV_HEADERS,
    YOUTUBE_VIDEO_SNIPPET_CSV_HEADERS,
    YOUTUBE_PLAYLIST_VIDEO_SNIPPET_CSV_HEADERS,
    YOUTUBE_COMMENT_CSV_HEADERS,
    YOUTUBE_CHANNEL_CSV_HEADERS,
)

YouTubeVideo = namedrecord(
    "YouTubeVideo", YOUTUBE_VIDEO_CSV_HEADERS, boolean=["has_caption"]
)

YouTubeVideoSnippet = namedrecord(
    "YouTubeVideoSnippet", YOUTUBE_VIDEO_SNIPPET_CSV_HEADERS
)

YouTubePlaylistVideoSnippet = namedrecord(
    "YouTubePlaylistVideoSnippet", YOUTUBE_PLAYLIST_VIDEO_SNIPPET_CSV_HEADERS
)

YouTubeComment = namedrecord("YoutubeComment", YOUTUBE_COMMENT_CSV_HEADERS)

YouTubeChannel = namedrecord(
    "YoutubeChannel",
    YOUTUBE_CHANNEL_CSV_HEADERS,
    boolean=["moderate_comments", "made_for_kids"],
    plural=["topic_ids", "topic_categories", "topic_keywords", "keywords"],
)


def get_int(item, key):
    nb = item.get(key)

    if nb is not None:
        return int(nb)

    return nb


def format_video(item):
    snippet = item["snippet"]
    stats = item["statistics"]
    details = item["contentDetails"]

    row = YouTubeVideo(
        item["id"],
        snippet["publishedAt"],
        snippet["channelId"],
        snippet["title"],
        snippet["description"],
        snippet["channelTitle"],
        get_int(stats, "viewCount"),
        get_int(stats, "likeCount"),
        # get_int(stats, 'dislikeCount'),
        # get_int(stats, 'favoriteCount'),
        get_int(stats, "commentCount"),
        details["duration"],
        details["caption"] == "true",
    )

    return row


def format_video_snippet(item):
    snippet = item["snippet"]

    row = YouTubeVideoSnippet(
        item["id"]["videoId"],
        snippet["publishedAt"],
        snippet["channelId"],
        snippet["title"],
        snippet["description"],
        snippet["channelTitle"],
    )

    return row


def format_comment(item):
    meta = item["snippet"]
    snippet = getpath(item, ["snippet", "topLevelComment", "snippet"])

    row = YouTubeComment(
        meta["videoId"],
        item["id"],
        snippet["authorDisplayName"],
        getpath(snippet, ["authorChannelId", "value"]),
        snippet["textOriginal"],
        int(snippet["likeCount"]),
        snippet["publishedAt"],
        snippet["updatedAt"],
        int(meta["totalReplyCount"]),
        None,
    )

    return row


def format_reply(item, video_id=None):
    snippet = item["snippet"]

    row = YouTubeComment(
        video_id if video_id is not None else snippet["videoId"],
        item["id"],
        snippet["authorDisplayName"],
        getpath(snippet, ["authorChannelId", "value"]),
        snippet["textOriginal"],
        int(snippet["likeCount"]),
        snippet["publishedAt"],
        snippet["updatedAt"],
        None,
        snippet["parentId"],
    )

    return row


def format_playlist_item_snippet(item):
    snippet = item["snippet"]

    row = YouTubePlaylistVideoSnippet(
        snippet["resourceId"]["videoId"],
        snippet["publishedAt"],
        snippet["channelId"],
        snippet["title"],
        snippet["description"],
        snippet["channelTitle"],
        snippet["position"],
    )

    return row


def format_channel(item):
    snippet = item.get("snippet")
    statistics = item.get("statistics")
    topic_details = item.get("topicDetails", {})
    status = item.get("status")
    branding_settings = item.get("brandingSettings")

    topic_ids = topic_details.get("topicIds", [])
    topic_categories = topic_details.get("topicCategories", [])
    topic_keywords = [url.rsplit("/", 1)[1] for url in topic_categories]

    keywords = getpath(branding_settings, ["channel", "keywords"])
    if keywords:
        keywords = [
            keyword.strip() for keyword in keywords.split('"') if keyword.strip()
        ]

    row = YouTubeChannel(
        item.get("id"),
        snippet.get("title"),
        snippet.get("description"),
        snippet.get("customUrl"),
        snippet.get("publishedAt"),
        getpath(snippet, ["thumbnails", "high", "url"]),
        snippet.get("defaultLanguage"),
        snippet.get("country"),
        getpath(item, ["contentDetails", "relatedPlaylists", "uploads"]),
        statistics.get("viewCount"),
        statistics.get("hiddenSubscriberCount"),
        statistics.get("subscriberCount"),
        statistics.get("videoCount"),
        topic_ids,
        topic_categories,
        topic_keywords,
        status.get("privacyStatus"),
        status.get("madeForKids"),
        status.get("longUploadsStatus"),
        keywords,
        getpath(branding_settings, ["channel", "moderateComments"]),
        getpath(branding_settings, ["channel", "unsubscribedTrailer"]),
        getpath(branding_settings, ["image", "bannerExternalUrl"]),
    )

    return row



# =============================================================================
# Minet Tiktok Exceptions
# =============================================================================
#
from minet.exceptions import MinetError


class TiktokError(MinetError):
    pass


class TiktokPublicAPIInvalidResponseError(TiktokError):
    def __init__(self, url, status, data):
        super().__init__()
        self.url = url
        self.status = status
        self.data = data

    def __str__(self):
        return super().__str__() + ", Url: %s, Status: %i, Data: %s" % (
            self.url,
            self.status,
            self.data,
        )


class TiktokInvalidCookieError(TiktokError):
    pass



# =============================================================================
# Minet Mediacloud CLI Action
# =============================================================================
#
# Logic of the `mc` action.
#
from minet.cli.argparse import (
    command,
    subcommand,
    ConfigAction,
    OutputAction,
    SplitterType,
)

MEDIACLOUD_MEDIAS_SUBCOMMAND = subcommand(
    "medias",
    "minet.cli.mediacloud.medias",
    title="Minet Mediacloud Medias Command",
    description="""
        Retrieve metadata about a list of Mediacloud medias.
    """,
    variadic_input={"dummy_column": "media", "item_label": "Mediacloud media id"},
    arguments=[
        {
            "flag": "--feeds",
            "help": "If given, path of the CSV file listing media RSS feeds.",
            "action": OutputAction,
        }
    ],
)

MEDIACLOUD_SEARCH_SUBCOMMAND = subcommand(
    "search",
    "minet.cli.mediacloud.search",
    title="Minet Mediacloud Search Command",
    description="""
        Search stories on the Mediacloud platform.

        To learn how to compose more relevant queries, check out this guide:
        https://mediacloud.org/support/query-guide
    """,
    arguments=[
        {"name": "query", "help": "Search query."},
        {
            "flags": ["-c", "--collections"],
            "help": "List of collection ids to search, separated by commas.",
            "type": SplitterType(),
        },
        {
            "flags": ["--filter-query"],
            "help": "Solr filter query `fq` to use. Can be used to optimize some parts of the query.",
        },
        {
            "flags": ["-m", "--medias"],
            "help": "List of media ids to search, separated by commas.",
            "type": SplitterType(),
        },
        {
            "flag": "--publish-day",
            "help": 'Only search stories published on provided day (iso format, e.g. "2018-03-24").',
        },
        {
            "flag": "--publish-month",
            "help": 'Only search stories published on provided month (iso format, e.g. "2018-03").',
        },
        {
            "flag": "--publish-year",
            "help": 'Only search stories published on provided year (iso format, e.g. "2018").',
        },
        {
            "flag": "--skip-count",
            "help": "Whether to skip the first API call counting the number of posts for the progress bar.",
            "action": "store_true",
        },
    ],
)

# TODO: unnest into topic-stories one day
MEDIACLOUD_TOPIC_SUBCOMMAND = subcommand(
    "topic",
    "minet.cli.mediacloud.topic",
    title="Minet Mediacloud Topic Command",
    description="""
        Gather information and aggregated stats about pages and groups of
        the designated dashboard (indicated by a given token).
    """,
    subparsers={
        "help": "Topic action to perform.",
        "title": "topic_actions",
        "dest": "mc_topic_action",
        "commands": {
            "stories": {
                "title": "Minet Mediacloud Topic Stories Command",
                "description": "Retrieves the list of stories from a mediacloud topic.",
                "arguments": [
                    {"name": "topic_id", "help": "Id of the topic."},
                    {
                        "flag": "--media-id",
                        "help": "Return only stories belonging to the given media_ids.",
                    },
                    {
                        "flag": "--from-media-id",
                        "help": "Return only stories that are linked from stories in the given media_id.",
                    },
                ],
            }
        },
    },
)

MEDIACLOUD_COMMAND = command(
    "mediacloud",
    "minet.cli.mediacloud",
    aliases=["mc"],
    title="Minet Mediacloud Command",
    description="""
        Commands related to the MIT Mediacloud API v2.
    """,
    common_arguments=[
        {
            "flags": ["-t", "--token"],
            "help": 'Mediacloud API token (also called "key" sometimes).',
            "rc_key": ["mediacloud", "token"],
            "action": ConfigAction,
            "required": True,
        }
    ],
    subcommands=[
        MEDIACLOUD_MEDIAS_SUBCOMMAND,
        MEDIACLOUD_SEARCH_SUBCOMMAND,
        MEDIACLOUD_TOPIC_SUBCOMMAND,
    ],
)



# =============================================================================
# Minet Utils
# =============================================================================
#
# Miscellaneous helper function used throughout the library.
#
import re
import hashlib
import json
import time
import string
from random import uniform


def fuzzy_int(value):
    try:
        return int(value)
    except ValueError:
        return int(float(value))


def md5(string):
    h = hashlib.md5()
    h.update(string.encode())
    return h.hexdigest()


DOUBLE_QUOTES_RE = re.compile(r'"')


def fix_ensure_ascii_json_string(s):
    try:
        return json.loads('"%s"' % DOUBLE_QUOTES_RE.sub('\\"', s))
    except json.decoder.JSONDecodeError:
        return s


class PseudoFStringFormatter(string.Formatter):
    def get_field(self, field_name, args, kwargs):
        result = eval(field_name, None, kwargs)

        return result, None


def sleep_with_entropy(seconds, max_random_addendum):
    random_addendum = uniform(0, max_random_addendum)
    time.sleep(seconds + random_addendum)


def is_binary_mimetype(m: str) -> bool:
    if m.startswith("text/"):
        return False

    if not m.startswith("application/"):
        return True

    second_part = m.split("/", 1)[-1]

    return not (
        "json" in second_part
        or "html" in second_part
        or "xml" in second_part
        or "yaml" in second_part
        or "yml" in second_part
        or second_part == "x-httpd-php"
    )


NUMBER_RE = re.compile(r"\d+[\.,]?\d*[KM]?")


def clean_human_readable_numbers(text):
    match = NUMBER_RE.search(text)

    if match is None:
        return text

    approx_likes = match.group(0)

    if "K" in approx_likes:
        approx_likes = str(int(float(approx_likes[:-1]) * 10**3))

    elif "M" in approx_likes:
        approx_likes = str(int(float(approx_likes[:-1]) * 10**6))

    approx_likes = approx_likes.replace(",", "")
    approx_likes = approx_likes.replace(".", "")

    return approx_likes


def message_flatmap(*messages, sep=" ", end="\n"):
    return sep.join(
        end.join(m for m in message) if not isinstance(message, str) else message
        for message in messages
    )



# =============================================================================
# Minet Scraper Compilation
# =============================================================================
#
# Schemes related to scraper definition "compilation".
#
import itertools
import soupsieve
from io import StringIO

from minet.scrape.constants import EXTRACTOR_NAMES


def escape_string_as_literal(string):
    double_quotes = string.count('"')
    single_quotes = string.count("'")

    quote = '"' if single_quotes > double_quotes else "'"

    return quote + string.replace(quote, "\\" + quote) + quote


class CompilerContext(object):
    def __init__(
        self,
        printer,
        counter=None,
        level=0,
        var="root",
        container="main_value",
        parent=None,
    ):
        self.printer = printer
        self.counter = counter if counter is not None else itertools.count(0)

        self.level = level
        self.identifier = next(self.counter)
        self.container = container
        self.var = var
        self.parent = parent

        self.container_type = "var"
        self.container_key = None

    def descend(self):
        return CompilerContext(
            printer=self.printer,
            counter=self.counter,
            level=self.level + 1,
            var="element_%i" % self.identifier,
            container="value_%i" % self.identifier,
            parent=self,
        )

    def linger(self):
        context = CompilerContext(
            printer=self.printer,
            counter=self.counter,
            level=self.level,
            var="element_%i" % self.identifier,
            container="value_%i" % self.identifier,
            parent=self,
        )

        return context

    def print(self, string, **kwargs):
        self.printer(
            string,
            level=self.level,
            parent=self.parent.identifier,
            id=self.identifier,
            next=self.identifier + 1,
            **kwargs
        )

    def yield_to_parent(self, expr, **kwargs):
        if self.parent.container_type == "list":
            self.print("value_{parent}.append(%s)" % expr, **kwargs)
        elif self.parent.container_type == "dict":
            self.print(
                "value_{parent}[%s] = %s"
                % (escape_string_as_literal(self.parent.container_key), expr),
                **kwargs
            )
        else:
            self.print("value_{parent} = %s" % expr, **kwargs)


class FieldsNode(object):
    def __init__(self, definition):
        self.definition = definition

    def unwrap(self):
        return self.definition


def compile_scraper(definition, as_string=False):
    output = StringIO()

    def printer(string, level=0, **kwargs):
        if kwargs:
            string = string.format(**kwargs)

        # Note: from extraneous state above
        if level:
            string = ("  " * level) + string

        print(string, file=output)

    scope = {"soupsieve": soupsieve}

    printer("def scrape(root, context={}):")
    printer("  value_0 = None")
    printer("  element_1 = root")

    root_context = CompilerContext(printer)
    initial_context = root_context.descend()

    def recurse(node, context):
        # Default extraction
        if node is None or (isinstance(node, str) and node in EXTRACTOR_NAMES):
            context.yield_to_parent("element_{id}.get_text().strip()")
            return

        # Attribute
        if isinstance(node, str):
            context.yield_to_parent(
                "element_{id}.get({attr})", attr=escape_string_as_literal(node)
            )
            return

        # Fields
        if isinstance(node, FieldsNode):
            fields = node.unwrap()
            context.container_type = "dict"

            context.print("value_{id} = dict()")

            for k, spec in fields.items():
                context.container_key = k
                next_context = context.linger()
                context.print(
                    "element_{target} = element_{id}", target=next_context.identifier
                )
                recurse(spec, next_context)

            context.yield_to_parent("value_{id}")

            return

        # Iterating
        if "iterator" in node:
            context.container_type = "list"

            context.print(
                "elements_{id} = soupsieve.select({selector}, element_{id})",
                selector=escape_string_as_literal(node["iterator"]),
            )
            context.print("value_{id} = []")

            context.print("for element_{next} in elements_{id}:")

            if "fields" in node:
                recurse(FieldsNode(node["fields"]), context.descend())
            else:
                recurse(node.get("item"), context.descend())

            context.yield_to_parent("value_{id}")

    recurse(definition, initial_context)

    printer("  return value_0")

    # Only return string
    if as_string:
        return output.getvalue()

    # Execute in scope to create function
    exec(output.getvalue(), scope)

    return scope["scrape"]



# =============================================================================
# Minet Tiktok API Scraper
# =============================================================================
#
# Tiktok public API "scraper".
#
from ebbe import getpath
from urllib.parse import quote

from minet.utils import sleep_with_entropy
from minet.cookies import coerce_cookie_for_url_from_browser
from minet.web import (
    create_pool_manager,
    create_request_retryer,
    request,
    retrying_method,
)
from minet.tiktok.constants import (
    TIKTOK_DEFAULT_THROTTLE,
    TIKTOK_PUBLIC_API_DEFAULT_TIMEOUT,
    TIKTOK_URL,
    TIKTOK_MIN_TIME_RETRYER,
    TIKTOK_MAX_RANDOM_ADDENDUM,
)
from minet.tiktok.exceptions import (
    TiktokInvalidCookieError,
    TiktokPublicAPIInvalidResponseError,
)
from minet.tiktok.formatters import (
    format_video,
)


def forge_video_search_url(query, offset):
    url = (
        "https://www.tiktok.com/api/search/general/full/?aid=1988&keyword=%s&offset=%s"
        % (quote(query), offset)
    )

    return url


class TiktokAPIScraper(object):
    def __init__(self, cookie="firefox"):
        self.pool_manager = create_pool_manager(
            timeout=TIKTOK_PUBLIC_API_DEFAULT_TIMEOUT
        )

        cookie = coerce_cookie_for_url_from_browser(cookie, TIKTOK_URL)

        if not cookie:
            raise TiktokInvalidCookieError

        self.cookie = cookie
        self.retryer = create_request_retryer(
            min=TIKTOK_MIN_TIME_RETRYER,
        )

    @retrying_method()
    def request_json(self, url):
        headers = {"Cookie": self.cookie}

        response = request(
            url,
            pool_manager=self.pool_manager,
            spoof_ua=True,
            headers=headers,
            known_encoding="utf-8",
        )

        if response.status >= 400:
            raise TiktokPublicAPIInvalidResponseError(
                url, response.status, response.text()
            )

        sleep_with_entropy(TIKTOK_DEFAULT_THROTTLE, TIKTOK_MAX_RANDOM_ADDENDUM)

        return response.json()

    def search_videos(self, query):
        cursor = None

        while True:
            url = forge_video_search_url(query, cursor)

            data = self.request_json(url)

            item_list = data.get("data")

            if not item_list:
                break

            for item in item_list:
                if item["type"] != 1:
                    continue
                item = item["item"]
                yield format_video(item)

            has_next_page = getpath(data, ["has_more"])

            if has_next_page != 1:
                break

            cursor = getpath(data, ["cursor"])



# =============================================================================
# Minet Mediacloud Medias CLI Action
# =============================================================================
#
# Logic of the `mc medias` action.
#
import casanova

from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.mediacloud.utils import with_mediacloud_fatal_errors
from minet.mediacloud import MediacloudAPIClient
from minet.mediacloud.constants import (
    MEDIACLOUD_MEDIA_CSV_HEADER,
    MEDIACLOUD_FEED_CSV_HEADER,
)


@with_mediacloud_fatal_errors
@with_enricher_and_loading_bar(
    headers=MEDIACLOUD_MEDIA_CSV_HEADER[1:], title="Fetching medias", unit="medias"
)
def action(cli_args, enricher, loading_bar):
    added_headers = MEDIACLOUD_MEDIA_CSV_HEADER[1:]

    feeds_writer = None

    if cli_args.feeds:
        added_headers.append("feeds")
        feeds_writer = casanova.writer(
            cli_args.feeds, fieldnames=MEDIACLOUD_FEED_CSV_HEADER
        )

    client = MediacloudAPIClient(cli_args.token)

    for row, media_id in enricher.cells(cli_args.column, with_rows=True):
        result = client.media(media_id)
        result = result.as_csv_row()[1:]

        if cli_args.feeds:
            feeds = client.feeds(media_id)

            enricher.writerow(row, result + [len(feeds)])

            for feed in feeds:
                feeds_writer.writerow(feed.as_csv_row())
        else:
            enricher.writerow(row, result)

        loading_bar.advance()



# =============================================================================
# Minet Facebook Comments CLI Action
# =============================================================================
#
# Logic of the `fb comments` action.
#
from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.facebook.utils import with_facebook_fatal_errors
from minet.facebook import FacebookMobileScraper
from minet.facebook.constants import FACEBOOK_COMMENT_CSV_HEADERS
from minet.facebook.exceptions import FacebookInvalidTargetError


@with_facebook_fatal_errors
@with_enricher_and_loading_bar(
    headers=FACEBOOK_COMMENT_CSV_HEADERS,
    title="Scraping comments",
    unit="posts",
    nested=True,
    sub_unit="comments",
    stats=[
        {"name": "calls", "style": "info"},
        {"name": "q", "style": "info"},
        {"name": "replies", "style": "info"},
    ],
)
def action(cli_args, enricher, loading_bar):
    scraper = FacebookMobileScraper(cli_args.cookie, throttle=cli_args.throttle)

    for i, row, url in enricher.enumerate_cells(
        cli_args.column, with_rows=True, start=1
    ):
        with loading_bar.step(url):
            try:
                batches = scraper.comments(url, per_call=True, detailed=True)
            except FacebookInvalidTargetError:
                loading_bar.print(
                    "Given url (line %i) is probably not a Facebook group post: %s"
                    % (i, url)
                )
                continue

            for _, batch in batches:
                for comment in batch:
                    enricher.writerow(row, comment.as_csv_row())
                    loading_bar.nested_advance()

                # loading_bar.update(
                #     calls=details["calls"],
                #     replies=details["replies"],
                #     q=details["queue_size"],
                # )



# =============================================================================
# Minet Youtube Videos CLI Action
# =============================================================================
#
# Action reading an input CSV file line by line and retrieving metadata about
# the given Youtube videos using Google's APIs.
#
from operator import itemgetter

from minet.cli.utils import with_enricher_and_loading_bar
from minet.youtube import YouTubeAPIClient
from minet.youtube.constants import YOUTUBE_VIDEO_CSV_HEADERS


@with_enricher_and_loading_bar(
    headers=YOUTUBE_VIDEO_CSV_HEADERS, title="Retrieving videos", unit="videos"
)
def action(cli_args, enricher, loading_bar):
    client = YouTubeAPIClient(cli_args.key)

    iterator = enricher.cells(cli_args.column, with_rows=True)

    for (row, _), video in client.videos(iterator, key=itemgetter(1)):
        enricher.writerow(row, video.as_csv_row() if video else None)
        loading_bar.advance()



from casanova import ThreadSafeResumer

from minet.cli.argparse import command, BooleanAction
from minet.cli.constants import DEFAULT_CONTENT_FOLDER
from minet.cli.exceptions import InvalidArgumentsError

# TODO: lazyloading issue
from minet.constants import (
    COOKIE_BROWSERS,
    DEFAULT_THROTTLE,
    DEFAULT_FETCH_MAX_REDIRECTS,
    DEFAULT_RESOLVE_MAX_REDIRECTS,
)

COMMON_ARGUMENTS = [
    {
        "flag": "--domain-parallelism",
        "help": "Max number of urls per domain to hit at the same time.",
        "type": int,
        "default": 1,
    },
    {
        "flags": ["-g", "--grab-cookies"],
        "help": 'Whether to attempt to grab cookies from your computer\'s browser (supports "firefox", "chrome", "chromium", "opera" and "edge").',
        "choices": COOKIE_BROWSERS,
    },
    {
        "flags": ["-H", "--header"],
        "help": "Custom headers used with every requests.",
        "action": "append",
        "dest": "headers",
    },
    {
        "flag": "--insecure",
        "help": "Whether to allow ssl errors when performing requests or not.",
        "action": "store_true",
    },
    {
        "flags": ["-t", "--threads"],
        "help": "Number of threads to use.",
        "type": int,
        "default": 25,
    },
    {
        "flag": "--throttle",
        "help": "Time to wait - in seconds - between 2 calls to the same domain.",
        "type": float,
        "default": DEFAULT_THROTTLE,
    },
    {
        "flag": "--timeout",
        "help": "Maximum time - in seconds - to spend for each request before triggering a timeout. Defaults to ~30s.",
        "type": float,
    },
    {
        "flag": "--url-template",
        "help": "A template for the urls to fetch. Handy e.g. if you need to build urls from ids etc.",
    },
    {
        "flags": ["-X", "--request"],
        "help": "The http method to use. Will default to GET.",
        "dest": "method",
        "default": "GET",
    },
]


def resolve_fetch_arguments(cli_args):
    # If we are hitting a single url we enable contents_in_report by default
    if cli_args.has_dummy_csv and cli_args.contents_in_report is None:
        cli_args.contents_in_report = True

    if cli_args.dont_save:
        cli_args.contents_in_report = False

    if cli_args.contents_in_report and cli_args.compress:
        raise InvalidArgumentsError(
            "Cannot both --compress and get --contents-in-report!"
        )


FETCH_COMMAND = command(
    "fetch",
    "minet.cli.fetch.fetch",
    title="Minet Fetch Command",
    description="""
        Use multiple threads to fetch batches of urls from a CSV file. The
        command outputs a CSV report with additional metadata about the
        HTTP calls and will generally write the retrieved files in a folder
        given by the user.
    """,
    epilog="""
        Columns being added to the output:

        . "fetch_original_index": index of the line in the original file (the output will be
            arbitrarily ordered since multiple requests are performed concurrently).
        . "resolved_url": final resolved url (after templating & solving redirects).
        . "http_status": HTTP status code of the request, e.g. 200, 404, 503 etc.
        . "datetime_utc": datetime when the response was finished.
        . "fetch_error": an error code if anything went wrong when performing the request.
        . "filename": path to the downloaded file, relative to the folder given
            through -O/--output-dir.
        . "mimetype": detected mimetype of the requested file.
        . "encoding": detected encoding of the requested file if relevant.
        . "body_size": size of the downloaded document in bytes.
        . "body": if -c/--contents-in-report is set, will contain the
            downloaded text and the files won't be written to disk.

        --folder-strategy options:

        . "flat": default choice, all files will be written in the indicated
            content folder.

        . "prefix-x": e.g. "prefix-4", files will be written in folders
            having a name that is the first x characters of the file's name.
            This is an efficient way to partition content into folders containing
            roughly the same number of files if the file names are random (which
            is the case by default since md5 hashes will be used).

        . "hostname": files will be written in folders based on their url's
            full host name.

        . "normalized-hostname": files will be written in folders based on
            their url's hostname stripped of some undesirable parts (such as
            "www.", or "m." or "fr.", for instance).

        Examples:

        . Fetching a single url (can be useful when piping):
            $ minet fetch "https://www.lemonde.fr"

        . Fetching a batch of url from existing CSV file:
            $ minet fetch url file.csv > report.csv

        . CSV input from stdin (mind the `-`):
            $ xsv select url file.csv | minet fetch url -i - > report.csv

        . Dowloading files in specific output directory:
            $ minet fetch url file.csv -O html > report.csv
    """,
    resolve=resolve_fetch_arguments,
    resumer=ThreadSafeResumer,
    variadic_input={"dummy_column": "url"},
    arguments=[
        *COMMON_ARGUMENTS,
        {
            "flag": "--max-redirects",
            "help": "Maximum number of redirections to follow before breaking.",
            "type": int,
            "default": DEFAULT_FETCH_MAX_REDIRECTS,
        },
        {
            "flag": "--compress",
            "help": "Whether to compress the contents.",
            "action": "store_true",
        },
        {
            "flags": ["-c", "--contents-in-report", "-w", "--no-contents-in-report"],
            "help": "Whether to include retrieved contents, e.g. html, directly in the report and avoid writing them in a separate folder. This requires to standardize encoding and won't work on binary formats. Note that --contents-in-report is the default when no input file is given.",
            "dest": "contents_in_report",
            "action": BooleanAction,
        },
        {
            "flags": ["-D", "--dont-save"],
            "help": "Use not to write any downloaded file on disk.",
            "action": "store_true",
        },
        {
            "flags": ["-O", "--output-dir"],
            "help": "Directory where the fetched files will be written.",
            "default": DEFAULT_CONTENT_FOLDER,
        },
        {
            "flags": ["-f", "--filename"],
            "help": 'Name of the column used to build retrieved file names. Defaults to a md5 hash of final url. If the provided file names have no extension (e.g. ".jpg", ".pdf", etc.) the correct extension will be added depending on the file type.',
        },
        {
            "flag": "--filename-template",
            "help": "A template for the name of the fetched files.",
        },
        {
            "flag": "--folder-strategy",
            "help": "Name of the strategy to be used to dispatch the retrieved files into folders to alleviate issues on some filesystems when a folder contains too much files. Note that this will be applied on top of --filename-template. All of the strategies are described at the end of this help.",
            "default": "flat",
        },
        {
            "flag": "--keep-failed-contents",
            "help": "Whether to keep & write contents for failed (i.e. non-200) http requests.",
            "action": "store_true",
        },
        {
            "flag": "--standardize-encoding",
            "help": "Whether to systematically convert retrieved text to UTF-8.",
            "action": "store_true",
        },
        {
            "flag": "--only-html",
            "help": (
                "Only download urls if they are not unambiguously not html (e.g. if the url has the .pdf or .jpg extension etc.). "
                "Then, if the url was downloaded, it will only be written on disk if the http body actually looks like html."
            ),
            "action": "store_true",
        },
    ],
)

RESOLVE_COMMAND = command(
    "resolve",
    "minet.cli.fetch.fetch",
    title="Minet Resolve Command",
    description="""
        Use multiple threads to resolve batches of urls from a CSV file. The
        command outputs a CSV report with additional metadata about the
        HTTP calls and the followed redirections.
    """,
    epilog="""
        Columns being added to the output:

        . "original_index": index of the line in the original file (the output will be
            arbitrarily ordered since multiple requests are performed concurrently).
        . "resolved_url": final resolved url (after solving redirects).
        . "http_status": HTTP status code of the request, e.g. 200, 404, 503 etc.
        . "resolution_error": an error code if anything went wrong when performing the request.
        . "redirect_count": total number of redirections to reach the final url.
        . "redirect_chain": list of redirection types separated by "|".

        Examples:

        . Resolving a batch of url from existing CSV file:
            $ minet resolve url_column file.csv > report.csv

        . CSV input from stdin (mind the `-`):
            $ xsv select url_column file.csv | minet resolve url_column - > report.csv

        . Resolving a single url:
            $ minet resolve https://lemonde.fr
    """,
    resumer=ThreadSafeResumer,
    variadic_input={"dummy_column": "url"},
    arguments=[
        *COMMON_ARGUMENTS,
        {
            "flag": "--max-redirects",
            "help": "Maximum number of redirections to follow before breaking.",
            "type": int,
            "default": DEFAULT_RESOLVE_MAX_REDIRECTS,
        },
        {
            "flag": "--follow-meta-refresh",
            "help": "Whether to follow meta refresh tags. Requires to buffer part of the response body, so it will slow things down.",
            "action": "store_true",
        },
        {
            "flag": "--follow-js-relocation",
            "help": "Whether to follow typical JavaScript window relocation. Requires to buffer part of the response body, so it will slow things down.",
            "action": "store_true",
        },
        {
            "flag": "--infer-redirection",
            "help": "Whether to try to heuristically infer redirections from the urls themselves, without requiring a HTTP call.",
            "action": "store_true",
        },
        {
            "flag": "--canonicalize",
            "help": "Whether to extract the canonical url from the html source code of the web page if found. Requires to buffer part of the response body, so it will slow things down.",
            "action": "store_true",
        },
        {
            "flag": "--only-shortened",
            "help": "Whether to only attempt to resolve urls that are probably shortened.",
            "action": "store_true",
        },
    ],
)



from minet.constants import COOKIE_BROWSERS
from minet.instagram.exceptions import InstagramInvalidCookieError

from minet.cli.utils import with_fatal_errors


def fatal_errors_hook(cli_args, e):
    if isinstance(e, InstagramInvalidCookieError):
        if cli_args.cookie in COOKIE_BROWSERS:
            return 'Could not extract relevant cookie from "%s".' % cli_args.cookie

        return [
            "Relevant cookie not found.",
            "An Instagram authentication cookie is necessary to be able to scrape Facebook comments.",
            "Use the --cookie flag to choose a browser from which to extract the cookie or give your cookie directly.",
        ]


def with_instagram_fatal_errors(fn):
    return with_fatal_errors(fatal_errors_hook)(fn)
